/*************************************************************************
 Timer.h
 
 Wrapper for the high-resolution timer. Can't be used if the hi-res timer
 doesn't exist.
  
 Author   :   Dave Astle
 Date     :   2/1/2001

 Copyright (c) 2002 Myopic Rhino Games. All rights reserved.
*************************************************************************/

#ifndef __GLAUTIMER_H_INCLUDED__
#define __GLAUTIMER_H_INCLUDED__

#include <windows.h>
#ifndef HIPERFTIMER
  #include <mmsystem.h>
  #pragma comment(lib, "winmm.lib")
#endif

class CTimer
{
public:
  CTimer() {}
  ~CTimer() {}

/*************************************************************************
 Init()

 If the hi-res timer is present, the tick rate is stored and the function
 returns true. Otherwise, the function returns false, and the timer should
 not be used.
*************************************************************************/
bool Init()
{
#ifdef HIPERFTIMER
  if (!QueryPerformanceFrequency(&m_ticksPerSecond))
  {
    // system doesn't support hi-res timer
    return false;
  }
#endif

  Reset();
  return true;
} // end Init()


/*************************************************************************
 Init()

 If the hi-res timer is present, the tick rate is stored and the function
 returns true. Otherwise, the function returns false, and the timer should
 not be used.
*************************************************************************/
float GetElapsedSeconds(unsigned long elapsedFrames = 1)
{
#ifdef HIPERFTIMER
  static LARGE_INTEGER s_lastTime = m_startTime;
  LARGE_INTEGER currentTime;

  QueryPerformanceCounter(&currentTime);
  float seconds =  (float)(currentTime.QuadPart - s_lastTime.QuadPart) / (float)m_ticksPerSecond.QuadPart;
#else
  static DWORD s_lastTime = m_startTime;
  DWORD currentTime;

  currentTime = timeGetTime();
  float seconds =  ((float)currentTime - (float)s_lastTime) / (float)1000;
#endif


  // reset the timer
  s_lastTime = currentTime;

  return seconds;
} // end GetElapsedSeconds()


/***************************************************************************
 GetFPS()

 Returns the average frames per second over elapsedFrames, which defaults to
 one. If this is not called every frame, the client should track the number
 of frames itself, and reset the value after this is called.
***************************************************************************/
float GetFPS(unsigned long elapsedFrames = 1)
{
  m_totalFrames += elapsedFrames;

#ifdef HIPERFTIMER
  static LARGE_INTEGER s_lastTime = m_startTime;
  LARGE_INTEGER currentTime;

  QueryPerformanceCounter(&currentTime);
  float fps = (float)(elapsedFrames * m_ticksPerSecond.QuadPart) / (float)(currentTime.QuadPart - s_lastTime.QuadPart);
#else
  static DWORD s_lastTime = m_startTime;
  DWORD currentTime;

  currentTime = timeGetTime();
  float fps = (float)(elapsedFrames * 1000) / (float)(currentTime - s_lastTime);
#endif
  
  // reset the timer
  s_lastTime = currentTime;

  return fps;
} // end GetFPS


/*************************************************************************
 GetAggregateFPS()

 Returns the average fps since the last time Reset() was called. Note that
 this is meant to be used in conjunction with LockFPS or GetFPS.
 Otherwise, the function has no way of knowing how many frames have
 actually elapsed.
*************************************************************************/
float GetAggregateFPS()
{
#ifdef HIPERFTIMER
  LARGE_INTEGER currentTime;

  QueryPerformanceCounter(&currentTime);

  return (float)(m_totalFrames * m_ticksPerSecond.QuadPart) / (float)(currentTime.QuadPart - m_startTime.QuadPart);
#else
  DWORD currentTime;
  currentTime = timeGetTime();
  return (float)(m_totalFrames * 1000) / (float)(currentTime - m_startTime);
#endif
} // end GetAggregateFPS()


/*************************************************************************
 LockFPS()

 Used to lock the frame rate to a set amount. This will block until enough
 time has passed to ensure that the fps won't go over the requested
 amount. Note that this can only keep the fps from going above the
 specified level; it can still drop below it. It is assumed that if used,
 this function will be called every frame. The value returned is the
 instantaneous fps, which will be <= targetFPS.
*************************************************************************/
float LockFPS(unsigned char targetFPS)
{
  if (targetFPS == 0)
    targetFPS = 1;

  m_totalFrames++;

  float   fps;
#ifdef HIPERFTIMER
  static LARGE_INTEGER s_lastTime = m_startTime;
  LARGE_INTEGER currentTime;

  // delay to maintain a constant frame rate
  do {
    QueryPerformanceCounter(&currentTime);
    fps = (float)m_ticksPerSecond.QuadPart/(float)(currentTime.QuadPart - s_lastTime.QuadPart);
  } while (fps > (float)targetFPS);
#else
  static DWORD s_lastTime = m_startTime;
  DWORD currentTime;

  // delay to maintain a constant frame rate
  do {
    currentTime = timeGetTime();
    fps = (float)1000/(float)(currentTime - s_lastTime);
  } while (fps > (float)targetFPS);
#endif
  // reset the timer
  s_lastTime = m_startTime;

  return fps;
} // end LockFPS()


/*************************************************************************
 Reset()

 Restarts the timer
*************************************************************************/
void Reset()
{
#ifdef HIPERFTIMER
  QueryPerformanceCounter(&m_startTime);
#else
  m_startTime = timeGetTime();
#endif
  m_totalFrames = 0;
}


private:
#ifdef HIPERFTIMER
  LARGE_INTEGER m_startTime;
  LARGE_INTEGER m_ticksPerSecond;
#else
  DWORD         m_startTime;
#endif
  unsigned long m_totalFrames;
};

#endif // __GLAUTIMER_H_INCLUDED_
